@@ -1,19 +1,23 @@
-from typing import Dict
-from concurrent import futures
+from typing import Dict, List
 
+from distributed.client import Client  # type: ignore
 import click
-import grpc  # type: ignore
 
-from cicadad.core.scenario import Scenario, scenario_runner, user_runner
+from cicadad.core.scenario import (
+    Scenario,
+    test_runner,
+    scenario_runner,
+    user_scheduler,
+)
 from cicadad.util.constants import (
+    DEFAULT_CONTAINER_MODE,
     DEFAULT_CONTAINER_SERVICE_ADDRESS,
     DEFAULT_DATASTORE_ADDRESS,
     DEFAULT_CONTEXT_STRING,
     DEFAULT_DOCKER_NETWORK,
+    DEFAULT_KUBE_NAMESPACE,
 )
 from cicadad.util.context import decode_context
-from cicadad.server.hub import HubServer
-from cicadad.protos import hub_pb2_grpc
 
 
 class Engine:
@@ -36,44 +40,50 @@ class Engine:
 
     def run_test(
         self,
+        tags: List[str],
+        test_id: str,
         image: str,
         network: str,
+        namespace: str,
         datastore_address: str,
         container_service_address: str,
+        container_mode: str,
     ):
         """Startup function when test container is created. Starts hub server.
 
         Args:
+            tags: (List[str]): List of tags to filter scenarios by
+            test_id: ID of test to event back to client
             image (str): Image containing test code. Passed to scenarios
             network (str): Network to start scenarios in
+            namepace (str): Kube namespace to place jobs in
             datastore_address (str): Address of datastore client to receive scenario results
             container_service_address (str): Address of container service to start scenarios
+            container_mode (str): DOCKER or KUBE container modes
         """
-        server = grpc.server(futures.ThreadPoolExecutor())
-
-        hub_pb2_grpc.add_HubServicer_to_server(
-            HubServer(
-                scenarios=self.scenarios.values(),
-                image=image,
-                network=network,
-                datastore_address=datastore_address,
-                container_service_address=container_service_address,
-            ),
-            server,
+        test_runner(
+            self.scenarios.values(),
+            list(tags),
+            test_id,
+            image,
+            network,
+            namespace,
+            datastore_address,
+            container_service_address,
+            container_mode,
         )
 
-        server.add_insecure_port("[::]:50051")
-        server.start()
-        server.wait_for_termination()
-
     def run_scenario(
         self,
         scenario_name: str,
+        test_id: str,
         scenario_id: str,
         image: str,
         network: str,
+        namespace: str,
         datastore_address: str,
         container_service_address: str,
+        container_mode: str,
         encoded_context: str,
     ):
         """Startup function when scenario is started. Begins running scenario's
@@ -81,11 +91,14 @@ class Engine:
 
         Args:
             scenario_name (str): Name of scenario being run
+            test_id (str): Test ID to send results to
             scenario_id (str): ID of scenario run
             image (str): Image containing test code
             network (str): Network to start users in
+            namespace (str): Namespace to place kube jobs in
             datastore_address (str): Address of datastore client to save and receive results
             container_service_address (str): Address of container service to start and stop users
+            container_mode (str): DOCKER or KUBE container modes
             encoded_context (str): Context from test containing previous results
         """
         scenario = self.scenarios[scenario_name]
@@ -93,18 +106,21 @@ class Engine:
 
         scenario_runner(
             scenario,
+            test_id,
             image,
             network,
+            namespace,
             scenario_id,
             datastore_address,
             container_service_address,
+            container_mode,
             context,
         )
 
     def run_user(
         self,
         scenario_name: str,
-        user_id: str,
+        user_manager_id: str,
         datastore_address: str,
         encoded_context: str,
     ):
@@ -112,16 +128,18 @@ class Engine:
 
         Args:
             scenario_name (str): Name of scenario being run
-            user_id (str): Unique ID of user assigned by scenario
+            user_manager_id (str): Unique ID of user manager assigned by scenario
             datastore_address (str): Address of datastore client to receive work and save results
             encoded_context (str): Context from test containing previous results
         """
         scenario = self.scenarios[scenario_name]
         context = decode_context(encoded_context)
+        client = Client()
 
-        user_runner(
+        user_scheduler(
+            client,
             scenario,
-            user_id,
+            user_manager_id,
             datastore_address,
             context,
         )
@@ -135,33 +153,54 @@ def engine_cli(ctx):
 
 @engine_cli.command()
 @click.pass_context
+@click.option("--tag", "-t", type=str, multiple=True, default=[])
+@click.option("--test-id", type=str, required=True)
 @click.option("--image", type=str, required=True)
 @click.option("--network", type=str, default=DEFAULT_DOCKER_NETWORK)
+@click.option("--namespace", type=str, default=DEFAULT_KUBE_NAMESPACE)
 @click.option("--datastore-address", type=str, default=DEFAULT_DATASTORE_ADDRESS)
 @click.option(
     "--container-service-address", type=str, default=DEFAULT_CONTAINER_SERVICE_ADDRESS
 )
-def run_test(ctx, image, network, datastore_address, container_service_address):
+@click.option("--container-mode", type=str, default=DEFAULT_CONTAINER_MODE)
+def run_test(
+    ctx,
+    tag,
+    test_id,
+    image,
+    network,
+    namespace,
+    datastore_address,
+    container_service_address,
+    container_mode,
+):
     engine: Engine = ctx.obj
 
     engine.run_test(
+        tag,
+        test_id,
         image,
         network,
+        namespace,
         datastore_address,
         container_service_address,
+        container_mode,
     )
 
 
 @engine_cli.command()
 @click.pass_context
 @click.option("--name", type=str, required=True)
+@click.option("--test-id", type=str, required=True)
 @click.option("--scenario-id", type=str, required=True)
 @click.option("--image", type=str, required=True)
 @click.option("--network", type=str, default=DEFAULT_DOCKER_NETWORK)
+@click.option("--namespace", type=str, default=DEFAULT_KUBE_NAMESPACE)
 @click.option("--datastore-address", type=str, default=DEFAULT_DATASTORE_ADDRESS)
 @click.option(
     "--container-service-address", type=str, default=DEFAULT_CONTAINER_SERVICE_ADDRESS
 )
+@click.option("--container-mode", type=str, default=DEFAULT_CONTAINER_MODE)
 @click.option(
     "--encoded-context",
     type=str,
@@ -170,22 +209,28 @@ def run_test(ctx, image, network, datastore_address, container_service_address):
 def run_scenario(
     ctx,
     name,
+    test_id,
     scenario_id,
     image,
     network,
+    namespace,
     datastore_address,
     container_service_address,
+    container_mode,
     encoded_context,
 ):
     engine: Engine = ctx.obj
 
     engine.run_scenario(
         name,
+        test_id,
         scenario_id,
         image,
         network,
+        namespace,
         datastore_address,
         container_service_address,
+        container_mode,
         encoded_context,
     )
 
@@ -193,7 +238,7 @@ def run_scenario(
 @engine_cli.command()
 @click.pass_context
 @click.option("--name", type=str, required=True)
-@click.option("--user-id", type=str, required=True)
+@click.option("--user-manager-id", type=str, required=True)
 @click.option("--datastore-address", type=str, default=DEFAULT_DATASTORE_ADDRESS)
 @click.option(
     "--encoded-context",
@@ -203,7 +248,7 @@ def run_scenario(
 def run_user(
     ctx,
     name,
-    user_id,
+    user_manager_id,
     datastore_address,
     encoded_context,
 ):
@@ -211,7 +256,7 @@ def run_user(
 
     engine.run_user(
         name,
-        user_id,
+        user_manager_id,
         datastore_address,
         encoded_context,
     )
