@@ -1,4 +1,4 @@
-from typing import Any, Callable, Dict, List, Optional, Set
+from typing import Any, Callable, Dict, Iterable, List, Optional, Set
 from datetime import datetime, timedelta
 import uuid
 import time
@@ -7,9 +7,10 @@ import io
 import traceback
 
 from pydantic import BaseModel, Field
+from cicadad.util.constants import KUBE_CONTAINER_MODE
+from dask.distributed import Client, fire_and_forget, secede  # type: ignore
 
 from cicadad.util.context import encode_context
-from cicadad.core import containers
 from cicadad.services import datastore, container_service
 from cicadad.util import printing
 
@@ -26,11 +27,11 @@ class UserCommands(object):
         Args:
             scenario (Scenario): Scenario being run
             user_id (str): ID of current user
-            scenario_id (str): ID of current scenario
             datastore_address: Address of datastore client
         """
         self.scenario = scenario
         self.user_id = user_id
+        self.event_buffer: List[datastore.UserEvent] = []
         self.datastore_address = datastore_address
 
         self.available_work = 0
@@ -41,9 +42,16 @@ class UserCommands(object):
         Returns:
             bool: User is up
         """
-        # HACK: User is stopped by scenario, but may be useful to explicitly
-        # tell user to stop to allow for a graceful user-defined exit
-        return True
+        return self.get_events("STOP_USER") == []
+
+    def get_events(self, kind: str):
+        """Get events sent to user from scenario
+
+        Returns:
+            List[UserEvent]: List of user events
+            kind: type of event to retrieve
+        """
+        return datastore.get_user_events(self.user_id, kind, self.datastore_address)
 
     def has_work(self, timeout_ms: Optional[int] = 1000):
         """Check if user has remaining invocations
@@ -90,8 +98,8 @@ class UserCommands(object):
                 output, exception = None, e
 
                 if log_traceback:
-                    print("Exception traceback:")
-                    traceback.print_tb(e.__traceback__)
+                    print("Exception traceback:", file=buffer)
+                    traceback.print_tb(e.__traceback__, file=buffer)
 
         return output, exception, buffer.getvalue()
 
@@ -122,33 +130,44 @@ class ScenarioCommands(object):
     def __init__(
         self,
         scenario: "Scenario",
+        test_id: str,
         image: str,
         network: str,
+        namespace: str,
         scenario_id: str,
         datastore_address: str,
         container_service_address: str,
+        container_mode: str,
         context: dict,
     ):
         """Commands available to a scenario
 
         Args:
             scenario (Scenario): Scenario being run
+            test_id (str): ID of test being run, used to send results
             image (str): Docker image for scenario
             network (str): Docker network for scenario
+            namespace (str): Kube namespace to place jobs in
             scenario_id (str): ID of scenario run
             datastore_address (str): Address of datastore to pass to users
             container_service (str): Address of container service to start and stop users
+            container_mode (str): DOCKER or KUBE mode to create containers
             context (dict): Context data to pass to users
         """
         self.scenario = scenario
+        self.test_id = test_id
         self.image = image
         self.network = network
+        self.namespace = namespace
         self.datastore_address = datastore_address
         self.container_service_address = container_service_address
+        self.container_mode = container_mode
         self.context = context
 
         self.scenario_id = scenario_id
         self.user_ids: Set[str] = set()
+        self.user_locations: Dict[str, str] = {}
+        self.user_manager_counts: Dict[str, int] = {}
         self.num_users = 0
         self.buffered_work = 0
         self.aggregated_results = None
@@ -171,49 +190,114 @@ class ScenarioCommands(object):
         Args:
             n (int): Number of users to start
         """
-        # FEATURE: will need to determine whether docker or kube args
-        # FEATURE: attach additional container args to scenario
         if n < 0:
             raise ValueError("Must supply a positive number of users to start")
 
         if n == 0:
             return
 
-        # FEATURE: mount env and volumes for user
-        for _ in range(n):
-            user_id = f"user-{self.scenario.name}-{str(uuid.uuid4())[:8]}"
-            encoded_context = encode_context(self.context)
+        users_to_start: Dict[str, int] = {}
+        remaining_users = n
 
-            args = containers.DockerServerArgs(
-                image=self.image,
-                name=user_id,
-                command=[
-                    "run-user",
-                    "--name",
-                    self.scenario.name,
-                    "--user-id",
-                    user_id,
-                    "--datastore-address",
-                    self.datastore_address,
-                    "--encoded-context",
-                    encoded_context,
-                ],
-                labels=["cicada-distributed-user", self.scenario.name],
-                # env: Dict[str, str]={}
-                # volumes: Optional[List[Volume]]
-                # host_port: Optional[int]
-                # container_port: Optional[int]
-                network=self.network,
-                # create_network: bool=True
+        # Attempt to fill existing user managers
+        for user_manager in self.user_manager_counts:
+            if remaining_users < 1:
+                break
+
+            remaining_capacity = (
+                self.scenario.users_per_container
+                - self.user_manager_counts[user_manager]
             )
 
-            self.user_ids.add(user_id)
-            container_service.start_container(args, self.container_service_address)
+            users_for_manager = min(remaining_capacity, remaining_users)
 
-        self.num_users += n
+            users_to_start[user_manager] = users_for_manager
+            remaining_users -= users_for_manager
+
+        # start user managers for remaining users
+        while remaining_users > 0:
+            user_manager_id = f"user-manager-{str(uuid.uuid4())[:8]}"
+            encoded_context = encode_context(self.context)
+
+            # FEATURE: mount env for user + specify args in scenario
+            if self.container_mode == KUBE_CONTAINER_MODE:
+                container_service.start_kube_container(
+                    container_service.StartKubeContainerArgs(
+                        name=user_manager_id,
+                        # env={}
+                        labels={
+                            "type": "cicada-distributed-user",
+                            "scenario": self.scenario.name,
+                            "test": self.test_id,
+                        },
+                        image=self.image,
+                        command=[
+                            "run-user",
+                            "--name",
+                            self.scenario.name,
+                            "--user-manager-id",
+                            user_manager_id,
+                            "--datastore-address",
+                            self.datastore_address,
+                            "--encoded-context",
+                            encoded_context,
+                        ],
+                        namespace=self.namespace,
+                    ),
+                    self.container_service_address,
+                )
+            else:
+                container_service.start_docker_container(
+                    container_service.StartDockerContainerArgs(
+                        name=user_manager_id,
+                        # env={}
+                        labels={
+                            "type": "cicada-distributed-user",
+                            "scenario": self.scenario.name,
+                            "test": self.test_id,
+                        },
+                        image=self.image,
+                        command=[
+                            "run-user",
+                            "--name",
+                            self.scenario.name,
+                            "--user-manager-id",
+                            user_manager_id,
+                            "--datastore-address",
+                            self.datastore_address,
+                            "--encoded-context",
+                            encoded_context,
+                        ],
+                        network=self.network,
+                    ),
+                    self.container_service_address,
+                )
+
+            users_for_manager = min(self.scenario.users_per_container, remaining_users)
+
+            users_to_start[user_manager_id] = users_for_manager
+            remaining_users -= users_for_manager
+
+        # Signal user managers to start users and update counts
+        for user_manager, num_users in users_to_start.items():
+            user_ids = [f"user-{str(uuid.uuid4())[:8]}" for _ in range(num_users)]
+
+            self._send_user_event(user_manager, "START_USERS", {"IDs": user_ids})
+
+            for user_id in user_ids:
+                self.user_ids.add(user_id)
+                self.user_locations[user_id] = user_manager
+
+            if user_manager in self.user_manager_counts:
+                self.user_manager_counts[user_manager] += num_users
+            else:
+                self.user_manager_counts[user_manager] = num_users
+
+            self.num_users += num_users
 
         # If called add_work before start_users, flush the saved work to the
         # new group of users
+        # NOTE: may be possible to get rid of this mechanic
         if self.buffered_work > 0:
             self.add_work(self.buffered_work)
             self.buffered_work = 0
@@ -239,7 +323,24 @@ class ScenarioCommands(object):
             if remaining < 1:
                 break
 
-            container_service.stop_container(user_id, self.container_service_address)
+            location = self.user_locations[user_id]
+            self._send_user_event(user_id, "STOP_USER", {})
+
+            self.user_manager_counts[location] -= 1
+
+            if self.user_manager_counts[location] < 1:
+                if self.container_mode == KUBE_CONTAINER_MODE:
+                    container_service.stop_kube_container(
+                        location,
+                        namespace=self.namespace,
+                        address=self.container_service_address,
+                    )
+                else:
+                    container_service.stop_docker_container(
+                        location,
+                        address=self.container_service_address,
+                    )
+
             self.user_ids.remove(user_id)
             self.num_users -= 1
             remaining -= 1
@@ -257,6 +358,19 @@ class ScenarioCommands(object):
 
         datastore.distribute_work(n, list(self.user_ids), self.datastore_address)
 
+    def _send_user_event(self, user_id: str, kind: str, payload: dict):
+        datastore.add_user_event(user_id, kind, payload, self.datastore_address)
+
+    def send_user_events(self, kind: str, payload: dict):
+        """Send an event to all user in the user pool.
+
+        Args:
+            kind (str): Type of event
+            payload (dict): JSON dict to send to user
+        """
+        for user_id in self.user_ids:
+            self._send_user_event(user_id, kind, payload)
+
     def get_latest_results(
         self,
         timeout_ms: Optional[int] = 1000,
@@ -332,7 +446,9 @@ class ScenarioCommands(object):
     #     return
 
 
-def filter_scenarios_by_tag(scenarios: List["Scenario"], tags: List[str]):
+def filter_scenarios_by_tag(
+    scenarios: Iterable["Scenario"], tags: List[str]
+) -> List["Scenario"]:
     """Filter scenarios that have tags in the list of tags provided. Returns
     all scenarios if tag list is empty
 
@@ -344,77 +460,162 @@ def filter_scenarios_by_tag(scenarios: List["Scenario"], tags: List[str]):
         List[Scenario]: List of filtered scenarios
     """
     if tags == []:
-        return scenarios
+        return list(scenarios)
 
     return [s for s in scenarios if set(s.tags).intersection(set(tags)) != set()]
 
 
-class TestStatus(BaseModel):
-    type: str
-    scenario: str
-    message: str
-    context: str
-
-
-def test_runner(
-    scenarios: List["Scenario"],
+def start_scenario(
+    scenario: "Scenario",
+    scenario_id: str,
+    results: Dict[str, dict],
+    container_mode: str,
     image: str,
+    test_id: str,
     network: str,
+    namespace: str,
     datastore_address: str,
     container_service_address: str,
 ):
-    started: Dict[str, str] = {}
-    results: Dict[str, dict] = {}
+    container_id = f"scenario-{scenario_id}"
+    encoded_context = encode_context(results)
 
-    # Start scenarios with no dependencies
-    for scenario in scenarios:
-        if scenario.dependencies == []:
-            # FIXME: move to function
-            scenario_id = str(uuid.uuid4())[:8]
-            container_id = f"scenario-{scenario.name}-{scenario_id}"
-            encoded_context = encode_context(results)
-
-            args = containers.DockerServerArgs(
+    if container_mode == KUBE_CONTAINER_MODE:
+        container_service.start_kube_container(
+            container_service.StartKubeContainerArgs(
                 image=image,
                 name=container_id,
                 command=[
                     "run-scenario",
                     "--name",
                     scenario.name,
+                    "--test-id",
+                    test_id,
                     "--scenario-id",
                     scenario_id,
                     "--image",
                     image,
                     "--network",
                     network,
+                    "--namespace",
+                    namespace,
                     "--encoded-context",
                     encoded_context,
                     "--datastore-address",
                     datastore_address,
                     "--container-service-address",
                     container_service_address,
+                    "--container-mode",
+                    container_mode,
                 ],
-                labels=["cicada-distributed-scenario", scenario.name],
+                labels={
+                    "type": "cicada-distributed-scenario",
+                    "scenario": scenario.name,
+                    "test": test_id,
+                },
+                # env: Dict[str, str]={}
+                namespace=namespace,
+            ),
+            container_service_address,
+        )
+    else:
+        container_service.start_docker_container(
+            container_service.StartDockerContainerArgs(
+                image=image,
+                name=container_id,
+                command=[
+                    "run-scenario",
+                    "--name",
+                    scenario.name,
+                    "--test-id",
+                    test_id,
+                    "--scenario-id",
+                    scenario_id,
+                    "--image",
+                    image,
+                    "--network",
+                    network,
+                    "--namespace",
+                    namespace,
+                    "--encoded-context",
+                    encoded_context,
+                    "--datastore-address",
+                    datastore_address,
+                    "--container-service-address",
+                    container_service_address,
+                    "--container-mode",
+                    container_mode,
+                ],
+                labels={
+                    "type": "cicada-distributed-scenario",
+                    "scenario": scenario.name,
+                    "test": test_id,
+                },
                 # env: Dict[str, str]={}
-                # volumes: Optional[List[Volume]]
-                # host_port: Optional[int]
-                # container_port: Optional[int]
                 network=network,
-                # create_network: bool=True
-            )
+            ),
+            container_service_address,
+        )
 
-            container_service.start_container(args, container_service_address)
+    datastore.add_test_event(
+        test_id=test_id,
+        kind="SCENARIO_STARTED",
+        event=datastore.TestStatus(
+            scenario=scenario.name,
+            message=f"Started scenario: {scenario.name} ({container_id})",
+            context=json.dumps(results),
+        ),
+        address=datastore_address,
+    )
+
+
+def test_runner(
+    scenarios: Iterable["Scenario"],
+    tags: List[str],
+    test_id: str,
+    image: str,
+    network: str,
+    namespace: str,
+    datastore_address: str,
+    container_service_address: str,
+    container_mode: str,
+):
+    started: Dict[str, str] = {}
+    results: Dict[str, dict] = {}
+
+    valid_scenarios = filter_scenarios_by_tag(scenarios, tags)
+
+    datastore.add_test_event(
+        test_id=test_id,
+        kind="TEST_STARTED",
+        event=datastore.TestStatus(
+            message=f"Collected {len(valid_scenarios)} Scenario(s)",
+        ),
+        address=datastore_address,
+    )
+
+    # Start scenarios with no dependencies
+    for scenario in valid_scenarios:
+        if scenario.dependencies == []:
+            scenario_id = str(uuid.uuid4())[:8]
+
+            start_scenario(
+                scenario,
+                scenario_id,
+                results,
+                container_mode,
+                image,
+                test_id,
+                network,
+                namespace,
+                datastore_address,
+                container_service_address,
+            )
             started[scenario.name] = scenario_id
 
-            yield TestStatus(
-                type="SCENARIO_STARTED",
-                scenario=scenario.name,
-                message=f"Started scenario: {scenario.name} ({container_id})",
-                context=json.dumps(results),
-            )
-
     # listen to completed events and start scenarios with dependencies
-    while len(results) != len(scenarios):
+    # FEATURE: scenario timeout counter here as well
+    while len(results) != len(valid_scenarios):
         for scenario_name in [sn for sn in started if sn not in results]:
             # FIXME: move logic to eventing function
             # FEATURE: stream back metrics gathered from scenarios
@@ -426,61 +627,38 @@ def test_runner(
             if result is not None:
                 results[scenario_name] = result
 
-                yield TestStatus(
-                    type="SCENARIO_FINISHED",
-                    scenario=scenario_name,
-                    message=f"Finished Scenario: {scenario_name}",
-                    context=json.dumps(results),
+                datastore.add_test_event(
+                    test_id=test_id,
+                    kind="SCENARIO_FINISHED",
+                    event=datastore.TestStatus(
+                        scenario=scenario_name,
+                        message=f"Finished Scenario: {scenario_name}",
+                        context=json.dumps(results),
+                    ),
+                    address=datastore_address,
                 )
 
-        for scenario in [s for s in scenarios if s.name not in started]:
+        for scenario in [s for s in valid_scenarios if s.name not in started]:
             # FIXME: move filtering to function
             if all(
                 dep.name in results and results[dep.name]["exception"] is None
                 for dep in scenario.dependencies
             ):
                 scenario_id = str(uuid.uuid4())[:8]
-                container_id = f"scenario-{scenario.name}-{scenario_id}"
-                encoded_context = encode_context(results)
 
-                args = containers.DockerServerArgs(
-                    image=image,
-                    name=container_id,
-                    command=[
-                        "run-scenario",
-                        "--name",
-                        scenario.name,
-                        "--scenario-id",
-                        scenario_id,
-                        "--image",
-                        image,
-                        "--network",
-                        network,
-                        "--encoded-context",
-                        encoded_context,
-                        "--datastore-address",
-                        datastore_address,
-                        "--container-service-address",
-                        container_service_address,
-                    ],
-                    labels=["cicada-distributed-scenario", scenario.name],
-                    # env: Dict[str, str]={}
-                    # volumes: Optional[List[Volume]]
-                    # host_port: Optional[int]
-                    # container_port: Optional[int]
-                    network=network,
-                    # create_network: bool=True
+                start_scenario(
+                    scenario,
+                    scenario_id,
+                    results,
+                    container_mode,
+                    image,
+                    test_id,
+                    network,
+                    namespace,
+                    datastore_address,
+                    container_service_address,
                 )
-
-                container_service.start_container(args, container_service_address)
                 started[scenario.name] = scenario_id
-
-                yield TestStatus(
-                    type="SCENARIO_STARTED",
-                    scenario=scenario.name,
-                    message=f"Started scenario: {scenario.name} ({container_id})",
-                    context=json.dumps(results),
-                )
             elif all(dep.name in results for dep in scenario.dependencies):
                 # has all dependencies but some are failed
                 started[scenario.name] = str(uuid.uuid4())[:8]
@@ -494,44 +672,66 @@ def test_runner(
                     ).json()
                 )
 
-                yield TestStatus(
-                    type="SCENARIO_FINISHED",
-                    scenario=scenario.name,
-                    message=f"Skipped Scenario: {scenario.name}",
-                    context=json.dumps(results),
+                datastore.add_test_event(
+                    test_id=test_id,
+                    kind="SCENARIO_FINISHED",
+                    event=datastore.TestStatus(
+                        scenario=scenario.name,
+                        message=f"Skipped Scenario: {scenario.name}",
+                        context=json.dumps(results),
+                    ),
+                    address=datastore_address,
                 )
 
         # NOTE: maybe make this configurable or shorter?
         time.sleep(1)
 
+    datastore.add_test_event(
+        test_id=test_id,
+        kind="TEST_FINISHED",
+        event=datastore.TestStatus(
+            message=f"Finished running {len(valid_scenarios)} Scenario(s)",
+        ),
+        address=datastore_address,
+    )
+
 
 def scenario_runner(
     scenario: "Scenario",
+    test_id: str,
     image: str,
     network: str,
+    namespace: str,
     scenario_id: str,
     datastore_address: str,
     container_service_address: str,
+    container_mode: str,
     context: dict,
 ):
     """Set up scenario environment and run scenario. Capture output and exceptions
 
     Args:
         scenario (Scenario): Scenario being run
+        test_id (str): Test id to send results
         image (str): Image for scenario
         network (str): Network to run scenario containers in
+        namespace (str): Kube namespace to place jobs in
         scenario_id (str): ID generated for scenario run
         datastore_address (str): Address of datastore passed to users
         container_service (str): Address of container service to start and stop users
+        container_mode (str): DOCKER or KUBE container mode
         context (dict): Test context to pass to users
     """
     scenario_commands = ScenarioCommands(
         scenario,
+        test_id,
         image,
         network,
+        namespace,
         scenario_id,
         datastore_address,
         container_service_address,
+        container_mode,
         context,
     )
 
@@ -583,6 +783,42 @@ def scenario_runner(
     )
 
 
+def user_scheduler(
+    scheduler: Client,
+    scenario: "Scenario",
+    user_manager_id: str,
+    datastore_address: str,
+    context: dict,
+):
+    """Schedules users inside user manager on events from scenario
+
+    Args:
+        scheduler (Client): Dask client to start users
+        scenario (Scenario): User Scenario
+        user_manager_id (str): ID of this user manager
+        datastore_address (str): Address of datastore to recieve events from
+        context (dict): Test context
+    """
+    while True:
+        events = datastore.get_user_events(
+            user_manager_id, "START_USERS", datastore_address
+        )
+
+        for event in events:
+            for user_id in event.payload["IDs"]:
+                fut = scheduler.submit(
+                    user_runner,
+                    scenario=scenario,
+                    user_id=user_id,
+                    datastore_address=datastore_address,
+                    context=context,
+                    pure=False,
+                )
+
+                # NOTE: may be better waiting for all futures to finish
+                fire_and_forget(fut)
+
+
 def user_runner(
     scenario: "Scenario",
     user_id: str,
@@ -597,12 +833,15 @@ def user_runner(
         datastore_address (str): Address of datastore client to save results
         context (dict): Test context containing previous scenario results
     """
+    secede()
+
     user_commands = UserCommands(
         scenario,
         user_id,
         datastore_address,
     )
 
+    # FEATURE: report errors here back to scenario
     scenario.user_loop(user_commands, context)  # type: ignore
 
 
@@ -618,6 +857,7 @@ def while_has_work(polling_timeout_ms: int = 1000):
         while user_commands.is_up():
             if user_commands.has_work(polling_timeout_ms):
                 start = datetime.now()
+                # FEATURE: option to retry on failure
                 output, exception, logs = user_commands.run(context=context)
                 end = datetime.now()
                 user_commands.report_result(
@@ -631,11 +871,7 @@ def while_has_work(polling_timeout_ms: int = 1000):
 
 
 def while_alive():
-    """Run user if hasn't been shut down yet
-
-    Args:
-        polling_timeout_ms (UserCommands): Time to wait for work before cycling
-    """
+    """Run user if hasn't been shut down yet"""
 
     def closure(user_commands: UserCommands, context: dict):
         while user_commands.is_up():
@@ -913,7 +1149,7 @@ def ramp_users_to_threshold(
     return closure
 
 
-def basic_verification(latest_results: List[datastore.Result]):
+def basic_verification(latest_results: List[datastore.Result], include_logs=True):
     """Create error strings for each errored result.
 
     Format:
@@ -925,11 +1161,24 @@ def basic_verification(latest_results: List[datastore.Result]):
     Returns:
         List[str]: List of error strings derived from results
     """
-    return [
-        f"* {type(result.exception)}: {str(result.exception)}"
-        for result in latest_results
-        if result.exception is not None
-    ]
+    exception_strings = []
+
+    for result in latest_results:
+        if result.exception is None:
+            continue
+
+        if include_logs:
+            exception_strings.append(
+                f"""
+* {type(result.exception)}: {str(result.exception)}
+{result.logs}"""
+            )
+        else:
+            exception_strings.append(
+                f"* {type(result.exception)}: {str(result.exception)}"
+            )
+
+    return exception_strings
 
 
 UserLoopFn = Callable[[UserCommands, dict], None]
@@ -948,6 +1197,7 @@ class Scenario(BaseModel):
     result_aggregator: Optional[ResultAggregatorFn]
     result_verifier: Optional[ResultVerifierFn] = Field(basic_verification)
     output_transformer: Optional[OutputTransformerFn]
+    users_per_container: int = 50
     tags: List[str] = []
 
 
@@ -955,6 +1205,7 @@ Scenario.update_forward_refs()
 
 
 def load_stages(*stages: LoadModelFn):
+    # FEATURE: signal user loop that stage has changed from scenario commands
     def closure(scenario_commands: ScenarioCommands, context: dict):
         for stage in stages:
             stage(scenario_commands, context)
@@ -962,6 +1213,3 @@ def load_stages(*stages: LoadModelFn):
         scenario_commands.scale_users(0)
 
     return closure
-
-
-# FEATURE: signal user loop that stage has changed from scenario commands
