@@ -1,7 +1,13 @@
 import base64
 import json
 
+CICADA_VERSION = "1.3.0"
+
 DEFAULT_DOCKER_NETWORK = "cicada-distributed-network"
+DEFAULT_KUBE_NAMESPACE = "default"
+DOCKER_CONTAINER_MODE = "DOCKER"
+KUBE_CONTAINER_MODE = "KUBE"
+DEFAULT_CONTAINER_MODE = DOCKER_CONTAINER_MODE
 
 DEFAULT_CONTEXT_STRING = base64.b64encode(json.dumps({}).encode("ascii")).decode(
     "ascii"
